import os
import sys
import json
import logging
import paho.mqtt.client as mqtt

from dbrepo.api.exceptions import AuthenticationError

logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-6s %(message)s', level=logging.INFO,
                    stream=sys.stdout)


class MqttClient:
    """
    The MqttClient class for communicating with the DBRepo MQTT API to import data. All parameters can be set also \
    via environment variables, e.g. set endpoint with DBREPO_ENDPOINT. You can override the constructor parameters \
    with the environment variables.

    :param broker_host: The MQTT API host. Optional. Default: "localhost".
    :param broker_port: The MQTT API port. Optional. Default: 1883,
    :param username: The MQTT API username. Optional.
    :param password: The MQTT API password. Optional.
    """
    broker_host: str = None
    broker_port: int = 1883
    username: str = None
    password: str = None

    def __init__(self,
                 broker_host: str = 'localhost',
                 broker_port: int = 1883,
                 username: str = None,
                 password: str = None) -> None:
        self.broker_host = os.environ.get('MQTT_API_HOST', broker_host)
        self.broker_port = os.environ.get('MQTT_API_PORT', broker_port)
        self.username = os.environ.get('MQTT_API_USERNAME', username)
        self.password = os.environ.get('MQTT_API_PASSWORD', password)

    def publish(self, routing_key: str, data=dict) -> None:
        """
        Publishes data to a given exchange with the given routing key with a blocking connection.

        :param routing_key: The routing key.
        :param data: The data.
        """
        if self.username is None or self.password is None:
            raise AuthenticationError(f"Failed to perform request: authentication required")
        client = mqtt.Client()
        client.username_pw_set(self.username, self.password)
        client.connect(self.broker_host, self.broker_port, 60)
        client.loop_start()
        client.publish(routing_key, json.dumps(data))
        client.loop_stop()



import json
import unittest

import requests_mock
from pandas import DataFrame

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import View, ViewColumn, ColumnType, UserBrief
from dbrepo.api.exceptions import ForbiddenError, NotExistsError, MalformedError, AuthenticationError


class ViewUnitTest(unittest.TestCase):

    def test_get_views_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/view', json=[])
            # test
            response = RestClient().get_views(database_id=1)
            self.assertEqual([], response)

    def test_get_views_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [View(id=1,
                        name="Data",
                        internal_name="data",
                        database_id=1,
                        initial_view=False,
                        query="SELECT id FROM mytable WHERE deg > 0",
                        query_hash="94c74728b11a690e51d64719868824735f0817b7",
                        owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                        is_public=True,
                        is_schema_public=True,
                        columns=[ViewColumn(id=1,
                                            ord=0,
                                            name="id",
                                            internal_name="id",
                                            database_id=1,
                                            auto_generated=False,
                                            type=ColumnType.BIGINT,
                                            is_public=True,
                                            is_null_allowed=False)],
                        identifiers=[])]
            # mock
            mock.get('/api/database/1/view', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_views(database_id=1)
            self.assertEqual(exp, response)

    def test_get_views_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/view', status_code=404)
            # test
            try:
                response = RestClient().get_views(database_id=1)
            except NotExistsError:
                pass

    def test_get_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = View(id=3,
                       name="Data",
                       internal_name="data",
                       database_id=1,
                       initial_view=False,
                       query="SELECT id FROM mytable WHERE deg > 0",
                       query_hash="94c74728b11a690e51d64719868824735f0817b7",
                       owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                       is_public=True,
                       is_schema_public=True,
                       columns=[ViewColumn(id=1,
                                           ord=0,
                                           name="id",
                                           internal_name="id",
                                           database_id=1,
                                           auto_generated=False,
                                           type=ColumnType.BIGINT,
                                           is_public=True,
                                           is_null_allowed=False)],
                       identifiers=[])
            # mock
            mock.get('/api/database/1/view/3', json=exp.model_dump())
            # test
            response = RestClient().get_view(database_id=1, view_id=3)
            self.assertEqual(exp, response)

    def test_get_view_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/view/3', status_code=403)
            # test
            try:
                response = RestClient().get_view(database_id=1, view_id=3)
            except ForbiddenError:
                pass

    def test_get_views_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/view/3', status_code=404)
            # test
            try:
                response = RestClient().get_view(database_id=1, view_id=3)
            except NotExistsError:
                pass

    def test_create_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = View(id=3,
                       name="Data",
                       internal_name="data",
                       database_id=1,
                       initial_view=False,
                       query="SELECT id FROM mytable WHERE deg > 0",
                       query_hash="94c74728b11a690e51d64719868824735f0817b7",
                       owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                       is_public=True,
                       is_schema_public=True,
                       columns=[ViewColumn(id=1,
                                           ord=0,
                                           name="id",
                                           internal_name="id",
                                           database_id=1,
                                           auto_generated=False,
                                           type=ColumnType.BIGINT,
                                           is_public=True,
                                           is_null_allowed=False)],
                       identifiers=[])
            # mock
            mock.post('/api/database/1/view', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_view(database_id=1, name="Data", is_public=True, is_schema_public=True,
                                          query="SELECT id FROM mytable WHERE deg > 0")
            self.assertEqual(exp, response)

    def test_create_view_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/view', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_view(database_id=1, name="Data", is_public=True, is_schema_public=True,
                                              query="SELECT id FROM mytable WHERE deg > 0")
            except MalformedError:
                pass

    def test_create_view_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/view', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_view(database_id=1, name="Data", is_public=True, is_schema_public=True,
                                              query="SELECT id FROM mytable WHERE deg > 0")
            except ForbiddenError:
                pass

    def test_create_view_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/view', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_view(database_id=1, name="Data", is_public=True, is_schema_public=True,
                                              query="SELECT id FROM mytable WHERE deg > 0")
            except NotExistsError:
                pass

    def test_create_view_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/view', status_code=404)
            # test
            try:
                response = RestClient().create_view(database_id=1, name="Data", is_public=True, is_schema_public=True,
                                                    query="SELECT id FROM mytable WHERE deg > 0")
            except AuthenticationError:
                pass

    def test_delete_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/view/3', status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_view(database_id=1, view_id=3)

    def test_delete_view_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/view/3', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id=1, view_id=3)
            except MalformedError:
                pass

    def test_delete_view_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/view/3', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id=1, view_id=3)
            except ForbiddenError:
                pass

    def test_delete_view_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/view/3', status_code=403)
            # test
            try:
                RestClient().delete_view(database_id=1, view_id=3)
            except AuthenticationError:
                pass

    def test_get_view_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(json.dumps(exp))
            # mock
            mock.get('/api/database/1/view/3/data', json=json.dumps(exp))
            # test
            response = RestClient().get_view_data(database_id=1, view_id=3)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_view_data_dataframe_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(json.dumps(exp))
            # mock
            mock.get('/api/database/1/view/3/data', json=json.dumps(exp))
            # test
            response: DataFrame = RestClient().get_view_data(database_id=1, view_id=3)
            self.assertEqual(df.shape, response.shape)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_view_data_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/view/3/data', status_code=400)
            # test
            try:
                response = RestClient().get_view_data(database_id=1, view_id=3)
            except MalformedError:
                pass

    def test_get_view_data_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/view/3/data', status_code=403)
            # test
            try:
                response = RestClient().get_view_data(database_id=1, view_id=3)
            except ForbiddenError:
                pass

    def test_get_view_data_count_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = 844737
            # mock
            mock.head('/api/database/1/view/3/data', headers={'X-Count': str(exp)})
            # test
            response = RestClient().get_view_data_count(database_id=1, view_id=3)
            self.assertEqual(exp, response)

    def test_get_view_data_count_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/database/1/view/3/data', status_code=400)
            # test
            try:
                response = RestClient().get_view_data_count(database_id=1, view_id=3)
            except MalformedError:
                pass

    def test_get_view_data_count_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/database/1/view/3/data', status_code=403)
            # test
            try:
                response = RestClient().get_view_data_count(database_id=1, view_id=3)
            except ForbiddenError:
                pass


if __name__ == "__main__":
    unittest.main()



import os
from unittest import TestCase, mock, main

from dbrepo.RestClient import RestClient


class DatabaseUnitTest(TestCase):

    def test_constructor_succeeds(self):
        # test
        client = RestClient()
        self.assertEqual("http://localhost", client.endpoint)
        self.assertIsNone(client.username)
        self.assertIsNone(client.password)
        self.assertTrue(client.secure)

    @mock.patch.dict(os.environ, {
        "REST_API_ENDPOINT": "https://test.dbrepo.tuwien.ac.at",
        "REST_API_USERNAME": "foo",
        "REST_API_PASSWORD": "bar",
        "REST_API_SECURE": "false",
    })
    def test_constructor_environment_succeeds(self):
        # test
        client = RestClient()
        self.assertEqual("https://test.dbrepo.tuwien.ac.at", client.endpoint)
        self.assertEqual("foo", client.username)
        self.assertEqual("bar", client.password)
        self.assertFalse(client.secure)

    def test_constructor_credentials_succeeds(self):
        # test
        client = RestClient(username='admin', password='pass')
        self.assertEqual("http://localhost", client.endpoint)
        self.assertEqual('admin', client.username)
        self.assertEqual('pass', client.password)
        self.assertTrue(client.secure)


if __name__ == "__main__":
    main()



import inspect
import logging
import sys
import unittest
from logging.config import dictConfig
from math import floor

from yaml import safe_load

from dbrepo.api import dto

logging.addLevelName(level=logging.NOTSET, levelName='TRACE')
logging.basicConfig(level=logging.DEBUG)

# logging configuration
dictConfig({
    'version': 1,
    'formatters': {
        'default': {
            'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        },
        'simple': {
            'format': '[%(asctime)s] %(levelname)s: %(message)s',
        },
    },
    'handlers': {'console': {
        'class': 'logging.StreamHandler',
        'stream': 'ext://sys.stdout',
        'formatter': 'simple'  # default
    }},
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    }
})


class AnalyseUnitTest(unittest.TestCase):
    schemas = None
    models: [()] = []
    found: int = 0
    skipped: [str] = []

    def setUp(self):
        with open('../../.docs/.openapi/api.yaml') as fh:
            self.schemas = safe_load(fh)['components']['schemas']
            for name, obj in inspect.getmembers(sys.modules[dto.__name__]):
                self.found += 1
                if not inspect.isclass(obj):
                    continue
                if f'{name}Dto' not in self.schemas:
                    logging.debug(f'skip model {name}: OpenAPI schema definition {name}Dto not found')
                    self.skipped.append(f'{name}Dto')
                    continue
                self.models.append((name, obj))

    def build_model(self, name: str, obj: any, definition: any) -> dict:
        model_dict = dict()
        for property in definition['properties']:
            if 'example' in definition['properties'][property]:
                if '$ref' not in definition['properties'][property]:
                    model_dict[property] = definition['properties'][property]['example']
                    continue
                ref = definition['properties'][property]['$ref'][len('#/components/schemas/'):-3]
                # recursive call
                model_dict[property] = self.build_model(ref, self.get_model(ref), self.schemas[f'{ref}Dto'])
                continue
            if 'items' in definition['properties'][property]:
                if '$ref' not in definition['properties'][property]['items']:
                    continue
                ref = definition['properties'][property]['items']['$ref'][len('#/components/schemas/'):-3]
                # recursive call
                model_dict[property] = [self.build_model(ref, self.get_model(ref), self.schemas[f'{ref}Dto'])]
                continue
            if '$ref' in definition['properties'][property]:
                ref = definition['properties'][property]['$ref'][len('#/components/schemas/'):-3]
                # recursive call
                model_dict[property] = self.build_model(ref, self.get_model(ref), self.schemas[f'{ref}Dto'])
        return model_dict

    def get_model(self, ref: str):
        for name, obj in self.models:
            if name == ref:
                return obj
        return None

    def test_dtos_succeeds(self):
        logging.info(f'Found {self.found} model(s) in {dto.__name__}')
        for name, obj in self.models:
            logging.debug(f'building model: {name} against OpenAPI schema definition {name}Dto')
            model = obj(**self.build_model(name, obj, self.schemas[f'{name}Dto']))
        logging.warning(f'Unable to find {len(self.skipped)} OpenAPI schema definition(s): {self.skipped}')
        logging.info(f'Coverage: {floor((1 - len(self.skipped) / self.found) * 100)}%')
        pass



import json
import unittest

import requests_mock
from pandas import DataFrame

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Table, CreateTableConstraints, Column, Constraints, ColumnType, ConceptBrief, UnitBrief, \
    TableStatistics, ColumnStatistic, PrimaryKey, ColumnBrief, TableBrief, UserBrief
from dbrepo.api.exceptions import MalformedError, ForbiddenError, NotExistsError, NameExistsError, \
    AuthenticationError, ExternalSystemError


class TableUnitTest(unittest.TestCase):

    def test_create_table_succeeds(self):
        exp = TableBrief(id=2,
                         database_id=1,
                         name="Test",
                         description="Test Table",
                         internal_name="test",
                         owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                         is_versioned=True,
                         is_public=True,
                         is_schema_public=True)
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_table(database_id=1, name="Test", description="Test Table", columns=[],
                                           is_public=True, is_schema_public=True,
                                           constraints=CreateTableConstraints())
            self.assertEqual(exp, response)

    def test_create_table_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id=1, name="Test", description="Test Table", columns=[],
                                               is_public=True, is_schema_public=True,
                                               constraints=CreateTableConstraints())
            except MalformedError:
                pass

    def test_create_table_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id=1, name="Test", description="Test Table", columns=[],
                                               is_public=True, is_schema_public=True,
                                               constraints=CreateTableConstraints())
            except ForbiddenError:
                pass

    def test_create_table_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id=1, name="Test", description="Test Table", columns=[],
                                               is_public=True, is_schema_public=True,
                                               constraints=CreateTableConstraints())
            except NotExistsError:
                pass

    def test_create_table_name_exists_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table', status_code=409)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id=1, name="Test", description="Test Table", columns=[],
                                               is_public=True, is_schema_public=True,
                                               constraints=CreateTableConstraints())
            except NameExistsError:
                pass

    def test_create_table_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table', status_code=409)
            # test
            try:
                response = RestClient().create_table(database_id=1, name="Test", description="Test Table", columns=[],
                                                     is_public=True, is_schema_public=True,
                                                     constraints=CreateTableConstraints())
            except AuthenticationError:
                pass

    def test_get_tables_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/table', json=[])
            # test
            response = RestClient().get_tables(database_id=1)
            self.assertEqual([], response)

    def test_get_tables_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [TableBrief(id=2,
                              name="Test",
                              description="Test Table",
                              database_id=1,
                              internal_name="test",
                              is_public=True,
                              is_schema_public=True,
                              owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                              is_versioned=True)]
            # mock
            mock.get('/api/database/1/table', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_tables(database_id=1)
            self.assertEqual(exp, response)

    def test_get_table_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Table(id=2,
                        name="Test",
                        description="Test Table",
                        database_id=1,
                        internal_name="test",
                        owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                        is_versioned=True,
                        queue_name='test',
                        routing_key='dbrepo.test_database_1234.test',
                        is_public=True,
                        is_schema_public=True,
                        constraints=Constraints(uniques=[],
                                                foreign_keys=[],
                                                checks=[],
                                                primary_key=[PrimaryKey(id=1,
                                                                        table=TableBrief(id=2, database_id=1,
                                                                                         name='Other',
                                                                                         internal_name='other',
                                                                                         description=None,
                                                                                         is_versioned=True,
                                                                                         is_public=True,
                                                                                         is_schema_public=True,
                                                                                         owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                        column=ColumnBrief(id=1, table_id=2,
                                                                                           database_id=1,
                                                                                           name='id',
                                                                                           alias=None,
                                                                                           internal_name='id',
                                                                                           type=ColumnType.BIGINT))]),
                        columns=[Column(id=1,
                                        name="ID",
                                        ord=0,
                                        database_id=1,
                                        table_id=2,
                                        internal_name="id",
                                        auto_generated=True,
                                        is_primary_key=True,
                                        type=ColumnType.BIGINT,
                                        is_public=True,
                                        is_null_allowed=False)])
            # mock
            mock.get('/api/database/1/table/2', json=exp.model_dump())
            # test
            response = RestClient().get_table(database_id=1, table_id=2)
            self.assertEqual(exp, response)

    def test_get_table_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/table/2', status_code=403)
            # test
            try:
                response = RestClient().get_table(database_id=1, table_id=2)
            except ForbiddenError:
                pass

    def test_get_table_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/table/2', status_code=404)
            # test
            try:
                response = RestClient().get_table(database_id=1, table_id=2)
            except NotExistsError:
                pass

    def test_delete_table_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/2', status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_table(database_id=1, table_id=2)

    def test_delete_table_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/2', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id=1, table_id=2)
            except ForbiddenError:
                pass

    def test_delete_table_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/2', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id=1, table_id=2)
            except NotExistsError:
                pass

    def test_delete_table_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/2', status_code=404)
            # test
            try:
                RestClient().delete_table(database_id=1, table_id=2)
            except AuthenticationError:
                pass

    def test_get_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(json.dumps(exp))
            # mock
            mock.get('/api/database/1/table/9/data', json=json.dumps(exp))
            # test
            response = RestClient().get_table_data(database_id=1, table_id=9)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_table_data_dataframe_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(json.dumps(exp))
            # mock
            mock.get('/api/database/1/table/9/data', json=json.dumps(exp))
            # test
            response = RestClient().get_table_data(database_id=1, table_id=9)
            self.assertEqual(df.shape, response.shape)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_table_data_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/table/9/data', status_code=400)
            # test
            try:
                response = RestClient().get_table_data(database_id=1, table_id=9)
            except MalformedError:
                pass

    def test_get_table_data_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/table/9/data', status_code=403)
            # test
            try:
                response = RestClient().get_table_data(database_id=1, table_id=9)
            except ForbiddenError:
                pass

    def test_get_table_data_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/table/9/data', status_code=404)
            # test
            try:
                response = RestClient().get_table_data(database_id=1, table_id=9)
            except NotExistsError:
                pass

    def test_get_table_data_count_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = 2
            # mock
            mock.head('/api/database/1/table/9/data', headers={'X-Count': str(exp)})
            # test
            response = RestClient().get_table_data_count(database_id=1, table_id=9)
            self.assertEqual(exp, response)

    def test_get_table_data_count_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/database/1/table/9/data', status_code=400)
            # test
            try:
                response = RestClient().get_table_data_count(database_id=1, table_id=9)
            except MalformedError:
                pass

    def test_get_table_data_count_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/database/1/table/9/data', status_code=403)
            # test
            try:
                response = RestClient().get_table_data_count(database_id=1, table_id=9)
            except ForbiddenError:
                pass

    def test_get_table_data_count_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/database/1/table/9/data', status_code=404)
            # test
            try:
                response = RestClient().get_table_data_count(database_id=1, table_id=9)
            except NotExistsError:
                pass

    def test_get_table_data_count_not_countable_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/database/1/table/9/data', status_code=409)
            # test
            try:
                response = RestClient().get_table_data_count(database_id=1, table_id=9)
            except ExternalSystemError:
                pass

    def test_create_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table/9/data', status_code=201)
            # test
            client = RestClient(username="a", password="b")
            client.create_table_data(database_id=1, table_id=9,
                                     data={'name': 'Josiah', 'age': 45, 'gender': 'male'})

    def test_create_table_data_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table/9/data', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id=1, table_id=9,
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except MalformedError:
                pass

    def test_create_table_data_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table/9/data', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id=1, table_id=9,
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except ForbiddenError:
                pass

    def test_create_table_data_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/table/9/data', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id=1, table_id=9,
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except NotExistsError:
                pass

    def test_update_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/9/data', status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.update_table_data(database_id=1, table_id=9,
                                     data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                     keys={'id': 1})

    def test_update_table_data_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/9/data', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id=1, table_id=9,
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except MalformedError:
                pass

    def test_update_table_data_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/9/data', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id=1, table_id=9,
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except ForbiddenError:
                pass

    def test_update_table_data_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/9/data', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id=1, table_id=9,
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except NotExistsError:
                pass

    def test_delete_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/9/data', status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_table_data(database_id=1, table_id=9, keys={'id': 1})

    def test_delete_table_data_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/9/data', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id=1, table_id=9, keys={'id': 1})
            except MalformedError:
                pass

    def test_delete_table_data_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/9/data', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id=1, table_id=9, keys={'id': 1})
            except ForbiddenError:
                pass

    def test_delete_table_data_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/9/data', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id=1, table_id=9, keys={'id': 1})
            except NotExistsError:
                pass

    def test_delete_table_data_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/table/9/data', status_code=404)
            # test
            try:
                RestClient().delete_table_data(database_id=1, table_id=9, keys={'id': 1})
            except AuthenticationError:
                pass

    def test_update_table_column_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Column(id=1,
                         ord=0,
                         name="ID",
                         database_id=1,
                         table_id=2,
                         internal_name="id",
                         auto_generated=True,
                         is_primary_key=True,
                         type=ColumnType.BIGINT,
                         is_public=True,
                         concept=ConceptBrief(id=2,
                                              uri="http://dbpedia.org/page/Category:Precipitation",
                                              name="Precipitation"),
                         unit=UnitBrief(id=2,
                                        uri="http://www.wikidata.org/entity/Q119856947",
                                        name="liters per square meter"),
                         is_null_allowed=False)
            # mock
            mock.put('/api/database/1/table/2/column/1', json=exp.model_dump(), status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.update_table_column(database_id=1, table_id=2, column_id=1,
                                                  unit_uri="http://www.wikidata.org/entity/Q119856947",
                                                  concept_uri="http://dbpedia.org/page/Category:Precipitation")
            self.assertEqual(exp, response)

    def test_update_table_column_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/2/column/1', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id=1, table_id=2, column_id=1,
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except MalformedError:
                pass

    def test_update_table_column_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/2/column/1', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id=1, table_id=2, column_id=1,
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except ForbiddenError:
                pass

    def test_update_table_column_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/2/column/1', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id=1, table_id=2, column_id=1,
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except NotExistsError:
                pass

    def test_update_table_column_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/table/2/column/1', status_code=404)
            # test
            try:
                RestClient().update_table_column(database_id=1, table_id=2, column_id=1,
                                                 unit_uri="http://www.wikidata.org/entity/Q119856947",
                                                 concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except AuthenticationError:
                pass

    def test_analyse_table_statistics_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = TableStatistics(
                columns={"id": ColumnStatistic(val_min=1.0, val_max=9.0, mean=5.0, median=5.0, std_dev=2.73)})
            # mock
            mock.get('/api/analyse/database/1/table/2/statistics', json=exp.model_dump(), status_code=202)
            # test
            response = RestClient().analyse_table_statistics(database_id=1, table_id=2)
            self.assertEqual(exp, response)

    def test_analyse_table_statistics_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/analyse/database/1/table/2/statistics', status_code=400)
            # test
            try:
                RestClient().analyse_table_statistics(database_id=1, table_id=2)
            except MalformedError:
                pass

    def test_analyse_table_statistics_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/analyse/database/1/table/2/statistics', status_code=404)
            # test
            try:
                RestClient().analyse_table_statistics(database_id=1, table_id=2)
            except NotExistsError:
                pass


if __name__ == "__main__":
    unittest.main()
